#!/usr/bin/python

####################################################
# Cornell Wren Tag Base Station Code
# Code developed for a Raspberry Pi B+
# Written for ECE 5990: Embedded Linux, Fall 2015
# Code developed by:
#     Taylor Pritchard (tjp79)
#     Shiva Rajagopal (svr24)
####################################################

# IMPORTS
import serial
import re
import socket
import urllib2
import sys
import datetime
import threading
import time
import io

# for JSON requests
# import json
# import requests

from smtplib import SMTP as smtp
from smtplib import SMTPException
from email.MIMEText import MIMEText as text
####################################################
# GLOBAL VARS
####################################################

# Internet variables for sending data
hostname = '104.236.255.240'
port = 60010
s = socket.socket()

# Serial filesystem objects
gpsFile = '/dev/ttyUSB0'
gpsBaud = 9600
tagFile = '/dev/ttyACM0'
tagBaud = 9600

# Global internet connection
conn = True
# Global current Time
currentTime = "TimeTest"
currentDate = "DateTest"
currentLat = ""
currentLong = ""

# A buffer to store logs while the system is offline
logBuffer = []
logEmailBuffer= []

# A variable to know whether we should print the GPS data on startup or not
startup = 1

# Regex patterns for Tag and GPGGA strings
tagPattern = re.compile(r'(\d{1}|[A-F])+(\d{1}|[A-F])+(\d{1}|[A-F])+(\d{1}|[A-F])+(\d{1}|[A-F])+(\d{1}|[A-F])+(\d{1}|[A-F])+(\d{1}|[A-F])+(\d{1}|[A-F])+(\d{1}|[A-F])+(\d{1}|[A-F])+(\d{1}|[A-F])')
gpggaPattern = re.compile(r'(\$GPGGA),+[0-2]+[0-3]+[0-5]+(\d{1})+[0-5]+\d{1}.+\d{3},+\d{4}.+\d{4},+(N|S),+\d{5}.+\d{4},+(E|W),+[0-8],+(0|1)+[0-2],+\d{1,2}.+\d{1},+(M),+(0.0),+(M),,(0000)+\*+(\d{1})+(\d{1}|[A-F])')

# Mail header constants
sender = 'piBaseStation@cornell.edu'
receivers = ['svr24@cornell.edu']
message = ''

####################################################
# HELPER FUNCTIONS
####################################################

# Checks for an internet connection by trying to access server

def internet_connected():
  try:
    response = urllib2.urlopen('http://'+hostname,timeout=0.5)
    response.close()
    return True
  except: pass
  return False

def send_mail(msg, subj):
  m= text(msg)
  m['From'] = sender
  m['To'] = receivers[0]
  m['Subject'] = subj
  success = 0
  while True:
    try:
      serv = smtp('localhost', timeout=1)
      serv.sendmail(sender, receivers, m.as_string())
      serv.quit()
      print "Sent " + subj + " email successfully"
      return 1
    except:
      if (subj == 'Backlog'):
        print "Failure to send " + subj + "... Retrying"
      elif (subj == 'New Data'):
        print "Failed to send new data, adding to log buffer"
        logEmailBuffer.append(message)
        break

def send_to_server(pack):
  data_json = json.dumps(pack)
  payload = {'json_payload': data_json}
  r = requests.get(hostname, data=payload)


####################################################
# LOGGING FUNCTION
####################################################

def logData(inStr):
  global conn, message, logBuffer, logEmailBuffer, currentLat, currentLong
  # Assemble string to send to server
  dateStr = time.strftime('%b %d %Y')
  timeStr = time.strftime('%H:%M:%s')
  locStr = currentLat + ',' + currentLong

  # JSON Format for server
  dataPacket = {'Tag': inStr, 'DetectionTime': timeStr, 'DetectionDate': dateStr, 'DetectionLocation': locStr}
  # Text format for email
  sendStr = 'Tag: '+inStr + '\tDetection Time: '+ timeStr + '\tDetection Date: '+dateStr + '\tDetection Location: '+locStr
 
  # check internet connection
  prevInetStatus = conn
  conn = internet_connected()
  
  # Depending on internet connection, take action 
  # Note: Startup conditions assume internet

  # Case 1: We lost internet, but just got it back
  if (prevInetStatus == False and conn == True): # Regained connection
    print 'Internet is back'

#   Send backlog up to server
#   for pack in logBuffer:
#     send_to_server(pack)
    
    # Send backlog via email
    for string in logEmailBuffer:
      message = message + string +'\n'
    message = message + sendStr
    thread = threading.Thread(target=send_mail, args=(message, 'Backlog'))
    thread.start()
    message = ''

    # Send the most recent tag to server
    #send_to_server(inStr, dateStr, timeStr, locStr)

    # Write most recent tag to log
    log.write(sendStr)
    
    # Clear backlog buffer
    logBuffer = []

    # Confirm sending of backlog in txt file
    log.write('\n***LOGGED TO SERVER***\n')

  # Case 2: We had internet, but lost it.
  elif (prevInetStatus == True and conn == False):
    print "We lost internet"
    
    # Write to offline log and backlog buffer
    log.write(sendStr + '\n')
    logBuffer.append(dataPacket)
    logEmailBuffer.append(sendStr)

  # Case 3: We still don't have internet
  elif (conn == False):
    
    # Write to offline log and backlog buffer
    log.write(sendStr + '\n')
    logBuffer.append(dataPacket)
    logEmailBuffer.append(sendStr)

  # Case 4: We still have internet
  elif (conn == True): 
    
    # Send to server
    # send_to_server(dataPacket)
    
    # Send email
    message = message + sendStr
    #success = send_mail(message, 'New Data')
    thread = threading.Thread(target=send_mail, args=(message, 'New Data'))
    thread.start()
    message = ''

    # Write to offline log with a * to signify sending to server
    log.write(sendStr+'*\n')

 
####################################################
# STRING SCANNERS
####################################################

# Parses out tag, assembles a string, and logs to the server
def tagScan(inStr):

  # Use regex to match string
  pattMatch = tagPattern.match(inStr)
  
  if pattMatch == None:
    return
  else:
    matchStr = pattMatch.group()
    
    # Build string
    tag = 'Tag:' + matchStr[0:8]
    linkQ = 'Signal Strength: ' + matchStr[8:10]
    signalQ = 'Link Quality: ' + matchStr[10:]
    combinedStr = tag+'\t'+linkQ+'\t'+signalQ+'\n'

    # Send data to log
    logData(matchStr)

    # Return string
    print combinedStr
    return
    

# Parses out the GPS data, writes to the log with data if it's on startup, and updates system time
def gpsScan(inStr):
  global startup, currentLat, currentLong 
  # Check for a match with regex
  pattMatch = gpggaPattern.match(inStr)

  if pattMatch == None:
    return
  else:
    matchstr= pattMatch.group()

    # Build string
    splitStr = matchstr.split(',')
    timeStr = 'GPS Time: ' + splitStr[1]
    latStr  = 'Latitude: ' + splitStr[2] + splitStr[3]
    longStr = 'Longitude: ' + splitStr[4] + splitStr[5]
    qualStr = 'Fix Quality: ' + splitStr[6]
    combinedStr = timeStr+'\t'+latStr+'\t'+longStr+'\t'+qualStr+'\n'

    # If the system is starting up, log the time, location, and quality
    if (startup == 1):
      log.write( combinedStr )
      startup = 0
    
    # Set the current system time
    currentTime = splitStr[1]
    currentLat = splitStr[2]+splitStr[3]
    currentLong = splitStr[4]+splitStr[5]

    # Send the combined string abck
    print combinedStr
    return

####################################################
# UART COMMUNICATION
####################################################

def tag_uart_comm():
  data= tagSer.readline()
  if (data == ''):
    return
  cleaned = data[2:6] + data[7:11] + data[12:16]
  tag_parsed_str = tagScan(cleaned)

def gps_uart_comm():
  data = gpsSer.readline()
  if (data == ''):
    return
  gps_parsed_str = gpsScan(data)

####################################################
# MAIN ENTRY POINT
####################################################

if __name__ == "__main__":
  try:
    s.connect((hostname, port))
  except:
    conn = False
  log = open("log.txt", 'a')
  gpsSer = serial.Serial (gpsFile,timeout=0.1)
  print "Connected to GPS"
  gpsSer.baudrate=gpsBaud

  tagSer = serial.Serial (tagFile,timeout=0.1)
  print "Connected to tag reader"
  tagSer.baudrate=tagBaud
  
  try: 
    while True:
      gps_uart_comm()
      tag_uart_comm()
  except KeyboardInterrupt:
    s.close()
    log.close()
    gpsSer.close()
    tagSer.close()
    sys.exit(0)


